__all__ = (
    "IntRounded",
    "int_binned_stat",
    "smallest_signed_angle",
    "SchemaConverter",
    "HpInComcamFov",
    "HpInLsstFov",
    "hp_kd_tree",
    "match_hp_resolution",
    "TargetoO",
    "SimTargetooServer",
    "set_default_nside",
    "restore_scheduler",
    "warm_start",
    "empty_observation",
    "scheduled_observation",
    "gnomonic_project_toxy",
    "gnomonic_project_tosky",
    "raster_sort",
    "run_info_table",
    "inrange",
    "season_calc",
    "create_season_offset",
    "thetaphi2xyz",
    "xyz2thetaphi",
    "mean_azimuth",
)

import datetime
import os
import socket
import sqlite3

import healpy as hp
import matplotlib.path as mplPath
import numpy as np
import pandas as pd

import rubin_scheduler.version as rsVersion
from rubin_scheduler.utils import _build_tree, _hpid2_ra_dec, _xyz_from_ra_dec, xyz_angular_radius


def smallest_signed_angle(a1, a2):
    """
    via https://stackoverflow.com/questions/1878907/
    the-smallest-difference-between-2-angles
    """
    two_pi = 2.0 * np.pi
    x = a1 % two_pi
    y = a2 % two_pi
    a = (x - y) % two_pi
    b = (y - x) % two_pi
    result = b + 0
    alb = np.where(a < b)[0]
    result[alb] = -1.0 * a[alb]
    return result


def thetaphi2xyz(theta, phi):
    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(phi)
    return x, y, z


def xyz2thetaphi(x, y, z):
    phi = np.arccos(z)
    theta = np.arctan2(y, x)
    return theta, phi


def mean_azimuth(az, min_val=0.1):
    """Compute the mean azimuth value accounting for wrap

    Parameters
    ----------
    az : `array-like`
        The azimuths to average. Radians
    min_val : `float`
        A min cutoff to just use pi as the mean. Default 0.1. Radians
    """

    x = np.cos(az)
    y = np.sin(az)
    meanx = np.mean(x)
    meany = np.mean(y)
    angle = np.arctan2(meany, meanx)
    radius = np.sqrt(meanx**2 + meany**2)
    mid_az = angle % (2.0 * np.pi)
    if IntRounded(radius) < IntRounded(min_val):
        mid_az = np.pi
    return mid_az


class IntRounded:
    """
    Class to help force comparisons be made on scaled up integers,
    preventing machine precision issues cross-platforms

    Note that casting a NaN to an int can have different behaviours
    cross-platforms, so will throw an error if attempted.

    Parameters
    ----------
    inval : number-like thing
        Some number that we want to compare
    scale : float (1e5)
        How much to scale inval before rounding and converting to an int.
    """

    def __init__(self, inval, scale=1e5):
        if np.any(~np.isfinite(inval)):
            raise ValueError("IntRounded can only take finite values.")
        self.initial = inval
        self.value = np.round(inval * scale).astype(int)
        self.scale = scale

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value

    def __repr__(self):
        return str(self.initial)

    def __add__(self, other):
        out_scale = np.min([self.scale, other.scale])
        result = IntRounded(self.initial + other.initial, scale=out_scale)
        return result

    def __sub__(self, other):
        out_scale = np.min([self.scale, other.scale])
        result = IntRounded(self.initial - other.initial, scale=out_scale)
        return result

    def __mul__(self, other):
        out_scale = np.min([self.scale, other.scale])
        result = IntRounded(self.initial * other.initial, scale=out_scale)
        return result

    def __div__(self, other):
        out_scale = np.min([self.scale, other.scale])
        result = IntRounded(self.initial / other.initial, scale=out_scale)
        return result


def set_default_nside(nside=None):
    """
    Utility function to set a default nside value across the scheduler.

    XXX-there might be a better way to do this.

    Parameters
    ----------
    nside : int (None)
        A valid healpixel nside.
    """
    if not hasattr(set_default_nside, "nside"):
        if nside is None:
            nside = 32
        set_default_nside.nside = nside
    if nside is not None:
        set_default_nside.nside = nside
    return set_default_nside.nside


def restore_scheduler(observation_id, scheduler, observatory, in_obs, filter_sched=None, fast=True):
    """Put the scheduler and observatory in the state they were in.
    Handy for checking reward fucnction

    Parameters
    ----------
    observation_id : int
        The ID of the last observation that should be completed
    scheduler : rubin_scheduler.scheduler.scheduler object
        Scheduler object.
    observatory : rubin_scheduler.scheduler.observatory.Model_observatory
        The observaotry object
    in_obs : np.array or str
        Array of observations (formated like
        rubin_scheduler.scheduler.empty_observation). If a string,
        assumed to be a file and SchemaConverter is used to load it.
    filter_sched : rubin_scheduler.scheduler.scheduler object
        The filter scheduler. Note that we don't look up the official
        end of the previous night, so there is potential for the
        loaded filters to not match.
    fast : bool (True)
        If True, loads observations and passes them as an array to
        the `add_observations_array` method. If False,
        passes observations individually with `add_observation` method.
    """
    if isinstance(in_obs, str):
        sc = SchemaConverter()
        # load up the observations
        observations = sc.opsim2obs(in_obs)
    else:
        observations = in_obs
    good_obs = np.where(observations["ID"] <= observation_id)[0]
    observations = observations[good_obs]

    # replay the observations back into the scheduler
    # In the future, may be able to replace this with a
    # faster .add_observations_array method.

    if fast:
        scheduler.add_observations_array(observations)
        obs = observations[-1]
    else:
        for obs in observations:
            scheduler.add_observation(obs)

    if filter_sched is not None:
        # We've assumed the filter scheduler doesn't have any filters
        # May need to call the add_observation method on it if that
        # changes.

        # Make sure we have mounted the right filters for the night
        # XXX--note, this might not be exact, but should work most
        # of the time.
        mjd_start_night = np.min(observations["mjd"][np.where(observations["night"] == obs["night"])])
        observatory.mjd = mjd_start_night
        conditions = observatory.return_conditions()
        filters_needed = filter_sched(conditions)
    else:
        filters_needed = ["u", "g", "r", "i", "y"]

    # update the observatory
    observatory.mjd = obs["mjd"] + observatory.observatory.visit_time(obs) / 3600.0 / 24.0
    observatory.obs_id_counter = obs["ID"] + 1
    observatory.observatory.parked = False
    observatory.observatory.current_ra_rad = obs["RA"]
    observatory.observatory.current_dec_rad = obs["dec"]
    observatory.observatory.current_rot_sky_pos_rad = obs["rotSkyPos"]
    observatory.observatory.cumulative_azimuth_rad = obs["cummTelAz"]
    observatory.observatory.current_filter = obs["filter"]
    observatory.observatory.mounted_filters = filters_needed
    # Note that we haven't updated last_az_rad, etc, but those
    # values should be ignored.

    return scheduler, observatory


def int_binned_stat(ids, values, statistic=np.mean):
    """
    Like scipy.binned_statistic, but for unique int ids
    """

    uids = np.unique(ids)
    order = np.argsort(ids)

    ordered_ids = ids[order]
    ordered_values = values[order]

    left = np.searchsorted(ordered_ids, uids, side="left")
    right = np.searchsorted(ordered_ids, uids, side="right")

    stat_results = []
    for le, ri in zip(left, right):
        stat_results.append(statistic(ordered_values[le:ri]))

    return uids, np.array(stat_results)


def gnomonic_project_toxy(ra1, dec1, r_acen, deccen):
    """Calculate x/y projection of ra1/dec1 in system with center
    at r_acen, deccen. Input radians. Grabbed from sims_selfcal"""
    # also used in Global Telescope Network website
    cosc = np.sin(deccen) * np.sin(dec1) + np.cos(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)
    x = np.cos(dec1) * np.sin(ra1 - r_acen) / cosc
    y = (np.cos(deccen) * np.sin(dec1) - np.sin(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)) / cosc
    return x, y


def gnomonic_project_tosky(x, y, r_acen, deccen):
    """Calculate RA/dec on sky of object with x/y and RA/Cen of
    field of view. Returns Ra/dec in radians."""
    denom = np.cos(deccen) - y * np.sin(deccen)
    RA = r_acen + np.arctan2(x, denom)
    dec = np.arctan2(np.sin(deccen) + y * np.cos(deccen), np.sqrt(x * x + denom * denom))
    return RA, dec


def match_hp_resolution(in_map, nside_out, unseen2nan=True):
    """Utility to convert healpix map resolution if needed and
    change hp.UNSEEN values to np.nan.

    Parameters
    ----------
    in_map : np.array
        A valie healpix map
    nside_out : int
        The desired resolution to convert in_map to
    unseen2nan : bool (True)
        If True, convert any hp.UNSEEN values to np.nan
    """
    current_nside = hp.npix2nside(np.size(in_map))
    if current_nside != nside_out:
        out_map = hp.ud_grade(in_map, nside_out=nside_out)
    else:
        out_map = in_map
    if unseen2nan:
        out_map[np.where(out_map == hp.UNSEEN)] = np.nan
    return out_map


def raster_sort(x0, order=["x", "y"], xbin=1.0):
    """XXXX--depriciated, use tsp instead.

    Do a sort to scan a grid up and down. Simple starting guess
    to traveling salesman.

    Parameters
    ----------
    x0 : array
    order : list
        Keys for the order x0 should be sorted in.
    xbin : float (1.)
        The bin_size to round off the first coordinate into

    returns
    -------
    array sorted so that it rasters up and down.
    """
    coords = x0.copy()
    bins = np.arange(
        coords[order[0]].min() - xbin / 2.0,
        coords[order[0]].max() + 3.0 * xbin / 2.0,
        xbin,
    )
    # digitize my bins
    coords[order[0]] = np.digitize(coords[order[0]], bins)
    order1 = np.argsort(coords, order=order)
    coords = coords[order1]
    places_to_invert = np.where(np.diff(coords[order[-1]]) < 0)[0]
    if np.size(places_to_invert) > 0:
        places_to_invert += 1
        indx = np.arange(coords.size)
        index_sorted = np.zeros(indx.size, dtype=int)
        index_sorted[0 : places_to_invert[0]] = indx[0 : places_to_invert[0]]

        for i, inv_pt in enumerate(places_to_invert[:-1]):
            if i % 2 == 0:
                index_sorted[inv_pt : places_to_invert[i + 1]] = indx[inv_pt : places_to_invert[i + 1]][::-1]
            else:
                index_sorted[inv_pt : places_to_invert[i + 1]] = indx[inv_pt : places_to_invert[i + 1]]

        if np.size(places_to_invert) % 2 != 0:
            index_sorted[places_to_invert[-1] :] = indx[places_to_invert[-1] :][::-1]
        else:
            index_sorted[places_to_invert[-1] :] = indx[places_to_invert[-1] :]
        return order1[index_sorted]
    else:
        return order1


class SchemaConverter:
    """
    Record how to convert an observation array to the standard
    opsim schema
    """

    def __init__(self):
        # Conversion dictionary, keys are opsim schema, values
        # are observation dtype names
        self.convert_dict = {
            "observationId": "ID",
            "night": "night",
            "observationStartMJD": "mjd",
            "observationStartLST": "lmst",
            "numExposures": "nexp",
            "visitTime": "visittime",
            "visitExposureTime": "exptime",
            "proposalId": "survey_id",
            "fieldId": "field_id",
            "fieldRA": "RA",
            "fieldDec": "dec",
            "altitude": "alt",
            "azimuth": "az",
            "filter": "filter",
            "airmass": "airmass",
            "skyBrightness": "skybrightness",
            "cloud": "clouds",
            "seeingFwhm500": "FWHM_500",
            "seeingFwhmGeom": "FWHM_geometric",
            "seeingFwhmEff": "FWHMeff",
            "fiveSigmaDepth": "fivesigmadepth",
            "slewTime": "slewtime",
            "slewDistance": "slewdist",
            "paraAngle": "pa",
            "psudoParaAngle": "psudo_pa",
            "rotTelPos": "rotTelPos",
            "rotTelPos_backup": "rotTelPos_backup",
            "rotSkyPos": "rotSkyPos",
            "rotSkyPos_desired": "rotSkyPos_desired",
            "moonRA": "moonRA",
            "moonDec": "moonDec",
            "moonAlt": "moonAlt",
            "moonAz": "moonAz",
            "moonDistance": "moonDist",
            "moonPhase": "moonPhase",
            "sunAlt": "sunAlt",
            "sunAz": "sunAz",
            "solarElong": "solarElong",
            "note": "note",
            "target": "target",
        }
        # Column(s) not bothering to remap:
        # 'observationStartTime': None,
        self.inv_map = {v: k for k, v in self.convert_dict.items()}
        # angles to convert
        self.angles_rad2deg = [
            "fieldRA",
            "fieldDec",
            "altitude",
            "azimuth",
            "slewDistance",
            "paraAngle",
            "psudoParaAngle",
            "rotTelPos",
            "rotSkyPos",
            "rotSkyPos_desired",
            "rotTelPos_backup",
            "moonRA",
            "moonDec",
            "moonAlt",
            "moonAz",
            "moonDistance",
            "sunAlt",
            "sunAz",
            "sunRA",
            "sunDec",
            "solarElong",
            "cummTelAz",
        ]
        # Put LMST into degrees too
        self.angles_hours2deg = ["observationStartLST"]

    def obs2opsim(self, obs_array, filename=None, info=None, delete_past=False, if_exists="append"):
        """Convert an array of observations into a pandas dataframe
        with Opsim schema.
        Parameters
        ----------
        obs_array : `np.array`
            Numpy array with OpSim observations.
        filename : `str`, optional
            Name of the database file to write to.
        info : `np.array`, optional
            Numpy array with database info.
        delete_past : `bool`
            Delete past observations (default=False)?
        if_exists : `str`
            Flag to pass to `to_sql` when writting to the
            database to control strategy when the database
            already exists.
        Returns
        -------
        `pd.DataFrame` or `None`
            Either the converted dataframe or `None`, if
            filename is provided.
        """
        if delete_past:
            try:
                os.remove(filename)
            except OSError:
                pass

        df = pd.DataFrame(obs_array)
        df = df.rename(index=str, columns=self.inv_map)
        for colname in self.angles_rad2deg:
            df[colname] = np.degrees(df[colname])
        for colname in self.angles_hours2deg:
            df[colname] = df[colname] * 360.0 / 24.0

        if filename is not None:
            con = sqlite3.connect(filename)
            df.to_sql("observations", con, index=False, if_exists=if_exists)
            if info is not None:
                df = pd.DataFrame(info)
                df.to_sql("info", con, if_exists=if_exists)
        else:
            return df

    def opsim2obs(self, filename):
        """convert an opsim schema dataframe into an observation array.
        Parameters
        ----------
        filename : `str`
            Sqlite file containing opsim output observations.
        """

        con = sqlite3.connect(filename)
        df = pd.read_sql("select * from observations;", con)
        for key in self.angles_rad2deg:
            try:
                df[key] = np.radians(df[key])
            except KeyError:
                df[key] = np.nan
        for key in self.angles_hours2deg:
            try:
                df[key] = df[key] * 24.0 / 360.0
            except KeyError:
                df[key] = np.nan

        df = df.rename(index=str, columns=self.convert_dict)

        blank = empty_observation()
        final_result = np.empty(df.shape[0], dtype=blank.dtype)
        # XXX-ugh, there has to be a better way.
        for key in df.columns:
            final_result[key] = df[key].values

        return final_result


def empty_observation(n=1):
    """Return a numpy array that could be a handy observation record

    Parameters
    ----------
    n : `int`
        Size of array to return. Default 1.

    Returns
    -------
    empty_observation : `np.array`

    The numpy fields have the following labels. These fields are
    required to be set to be a valid observation the model observatory
    can execute.
    RA : `float`
       The Right Acension of the observation (center of the field)
       (Radians)
    dec : `float`
       Declination of the observation (Radians)
    mjd : `float`
       Modified Julian Date at the start of the observation
       (time shutter opens)
    exptime : `float`
       Total exposure time of the visit (seconds)
    filter : `str`
        The filter used. Should be one of u, g, r, i, z, y.
    rotSkyPos : `float`
        The rotation angle of the camera relative to the sky E of N
        (Radians). Will be ignored if rotTelPos is finite.
        If rotSkyPos is set to NaN, rotSkyPos_desired is used.
    rotTelPos : `float`
        The rotation angle of the camera relative to the telescope
        (radians). Set to np.nan to force rotSkyPos to be used.
    rotSkyPos_desired : `float`
        If both rotSkyPos and rotTelPos are None/NaN, then
        rotSkyPos_desired (radians) is used. If rotSkyPos_desired
        results in a valid rotTelPos, rotSkyPos is set to
        rotSkyPos_desired. If rotSkyPos and rotTelPos are both NaN,
        and rotSkyPos_desired results in an out of range value for the
        camera rotator, then rotTelPos_backup is used.
    rotTelPos_backup : `float`
        Rotation angle of the camera relative to the telescope (radians).
        Only used as a last resort if rotSkyPos and rotTelPos are set
        to NaN and rotSkyPos_desired results in an out of range rotator
        value.
    nexp : `int`
        Number of exposures in the visit.
    flush_by_mjd : `float`
        If we hit this MJD, we should flush the queue and refill it.
    note : `str` (optional)
        Usually good to set the note field so one knows which survey
        object generated the observation.
    target : `str` (optional)
        A note about what target is being observed.

    Notes
    -----

    On the camera rotator angle. Order of priority goes:
    rotTelPos > rotSkyPos > rotSkyPos_desired > rotTelPos_backup
    where if rotTelPos is NaN, it checks rotSkyPos. If rotSkyPos is set,
    but not at an accessible rotTelPos, the observation will fail.
    If rotSkyPos is NaN, then rotSkyPos_desired is used. If
    rotSkyPos_desired is at an inaccessbile rotTelPos, the observation
    does not fail, but falls back to the value in rotTelPos_backup.

    Lots of additional fields that get filled in by the model observatory
    when the observation is completed.
    See documentation at:
    https://rubin-scheduler.lsst.io/output_schema.html

    """

    names = [
        "ID",
        "RA",
        "dec",
        "mjd",
        "flush_by_mjd",
        "exptime",
        "filter",
        "rotSkyPos",
        "rotSkyPos_desired",
        "nexp",
        "airmass",
        "FWHM_500",
        "FWHMeff",
        "FWHM_geometric",
        "skybrightness",
        "night",
        "slewtime",
        "visittime",
        "slewdist",
        "fivesigmadepth",
        "alt",
        "az",
        "pa",
        "psudo_pa",
        "clouds",
        "moonAlt",
        "sunAlt",
        "note",
        "target",
        "field_id",
        "survey_id",
        "block_id",
        "lmst",
        "rotTelPos",
        "rotTelPos_backup",
        "moonAz",
        "sunAz",
        "sunRA",
        "sunDec",
        "moonRA",
        "moonDec",
        "moonDist",
        "solarElong",
        "moonPhase",
        "cummTelAz",
        "scripted_id",
    ]

    types = [
        int,
        float,
        float,
        float,
        float,
        float,
        "U40",
        float,
        float,
        int,
        float,
        float,
        float,
        float,
        float,
        int,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        "U40",
        "U40",
        int,
        int,
        int,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        int,
    ]
    result = np.zeros(n, dtype=list(zip(names, types)))
    return result


def scheduled_observation(n=1):
    """Make an array to hold pre-scheduling observations

    Returns
    -------
    result : np.array


    Note
    ----
    mjd_tol : `float`
        The tolerance on how early an observation can execute (days).
        Observation will be considered valid to attempt
        when mjd-mjd_tol < current MJD < flush_by_mjd (and other
        conditions below pass)
    dist_tol : `float`
        The angular distance an observation can be away from the
        specified RA,Dec and still count as completing the observation
        (radians).
    alt_min : `float`
        The minimum altitude to consider executing the observation
        (radians).
    alt_max : `float`
        The maximuim altitude to try observing (radians).
    HA_max : `float`
        Hour angle limit. Constraint is such that for hour angle
        running from 0 to 24 hours, the target RA,Dec must be greather
        than HA_max and less than HA_min. Set HA_max to 0 for no
        limit. (hours)
    HA_min : `float`
        Hour angle limit. Constraint is such that for hour angle
        running from 0 to 24 hours, the target RA,Dec must be greather
        than HA_max and less than HA_min. Set HA_min to 24 for
        no limit. (hours)
    sun_alt_max : `float`
        The sun must be below sun_alt_max to execute. (radians)
    moon_min_distance : `float`
        The minimum distance to demand the moon should be away (radians)
    observed : `bool`
        If set to True, scheduler will probably consider this a
        completed observation and never attempt it.

    """

    # Standard things from the usual observations
    names = [
        "ID",
        "RA",
        "dec",
        "mjd",
        "flush_by_mjd",
        "exptime",
        "filter",
        "rotSkyPos",
        "rotTelPos",
        "rotTelPos_backup",
        "rotSkyPos_desired",
        "nexp",
        "note",
        "target",
    ]
    types = [
        int,
        float,
        float,
        float,
        float,
        float,
        "U1",
        float,
        float,
        float,
        float,
        int,
        "U40",
        "U40",
    ]
    names += [
        "mjd_tol",
        "dist_tol",
        "alt_min",
        "alt_max",
        "HA_max",
        "HA_min",
        "sun_alt_max",
        "moon_min_distance",
        "observed",
        "scripted_id",
    ]
    types += [float, float, float, float, float, float, float, float, bool, int]
    result = np.zeros(n, dtype=list(zip(names, types)))
    return result


def hp_kd_tree(nside=None, leafsize=100, scale=1e5):
    """
    Generate a KD-tree of healpixel locations

    Parameters
    ----------
    nside : int
        A valid healpix nside
    leafsize : int (100)
        Leafsize of the kdtree

    Returns
    -------
    tree : scipy kdtree
    """
    if nside is None:
        nside = set_default_nside()

    hpid = np.arange(hp.nside2npix(nside))
    ra, dec = _hpid2_ra_dec(nside, hpid)
    return _build_tree(ra, dec, leafsize, scale=scale)


class HpInLsstFov:
    """Return the healpixels in an underlying healpix grid that
    overlap an observation/pointing.

    This uses a very simple circular LSST camera model with no chip/raft gaps.

    Parameters
    ----------
    nside : `int`, optional
        Nside to match for the healpix array.
        Default None uses `set_default_nside`.
    fov_radius : `float`, optional
        Radius of the field of view in degrees. Default 1.75
        covers the inscribed circle.
    scale : `float`, optional
        How many sig figs to round when considering matches to healpixels.
        Useful for ensuring identical results cross-ploatform where
        float precision can vary.


    Examples
    --------
    Set up the class, then call to convert pointings to indices in the
    healpix array. Note that RA and dec should be in RADIANS.

    ```
    >>> ra = np.radians(30)
    >>> dec = np.radians(-20)
    >>> pointing2indx = HpInLsstFov()
    >>> indices = pointing2indx(ra, dec)
    >>> indices
    [8138, 8267]
    ```

    """

    def __init__(self, nside=None, fov_radius=1.75, scale=1e5):
        if nside is None:
            nside = set_default_nside()

        self.tree = hp_kd_tree(nside=nside, scale=scale)
        self.radius = np.round(xyz_angular_radius(fov_radius) * scale).astype(int)
        self.scale = scale

    def __call__(self, ra, dec, **kwargs):
        """
        Parameters
        ----------
        ra : float, array
            RA in radians
        dec : float, array
            Dec in radians

        Returns
        -------
        indx : numpy array
            The healpixels that are within the FoV
        """
        x, y, z = _xyz_from_ra_dec(ra, dec)
        x = np.round(x * self.scale).astype(int)
        y = np.round(y * self.scale).astype(int)
        z = np.round(z * self.scale).astype(int)

        if np.size(x) == 1:
            indices = self.tree.query_ball_point((np.max(x), np.max(y), np.max(z)), self.radius)
        else:
            indices = self.tree.query_ball_point(np.vstack([x, y, z]).T, self.radius)
        return indices


class HpInComcamFov:
    """
    Return the healpixels within a ComCam pointing. Simple camera model
    with no chip gaps.
    """

    def __init__(self, nside=None, side_length=0.7, scale=1e5):
        """
        Parameters
        ----------
        side_length : float (0.7)
            The length of one side of the square field of view (degrees).
        """
        if nside is None:
            nside = set_default_nside()
        self.nside = nside
        self.scale = scale
        self.tree = hp_kd_tree(nside=nside, scale=scale)
        self.side_length = np.round(np.radians(side_length * scale)).astype(int)
        self.inner_radius = np.round(xyz_angular_radius(side_length / 2.0) * scale).astype(int)
        self.outter_radius = np.round(xyz_angular_radius(side_length / 2.0 * np.sqrt(2.0)) * scale).astype(
            int
        )
        # The positions of the raft corners, unrotated
        self.corners_x = np.array(
            [
                -self.side_length / 2.0,
                -self.side_length / 2.0,
                self.side_length / 2.0,
                self.side_length / 2.0,
            ]
        )
        self.corners_y = np.array(
            [
                self.side_length / 2.0,
                -self.side_length / 2.0,
                -self.side_length / 2.0,
                self.side_length / 2.0,
            ]
        )

    def __call__(self, ra, dec, rotSkyPos=0.0):
        """
        Parameters
        ----------
        ra : float
            RA in radians
        dec : float
            Dec in radians
        rotSkyPos : float
            The rotation angle of the camera in radians
        Returns
        -------
        indx : numpy array
            The healpixels that are within the FoV
        """
        x, y, z = _xyz_from_ra_dec(ra, dec)
        x = np.round(x * self.scale).astype(int)
        y = np.round(y * self.scale).astype(int)
        z = np.round(z * self.scale).astype(int)
        # Healpixels within the inner circle
        indices = self.tree.query_ball_point((x, y, z), self.inner_radius)
        # Healpixels withing the outer circle
        indices_all = np.array(self.tree.query_ball_point((x, y, z), self.outter_radius))
        # Only need to check pixel if it is outside inner circle
        indices_to_check = indices_all[np.in1d(indices_all, indices, invert=True)]

        if np.size(indices_to_check) == 0:
            ValueError("No HEALpix in pointing. Maybe need to increase nside.")

        cos_rot = np.cos(rotSkyPos)
        sin_rot = np.sin(rotSkyPos)
        x_rotated = self.corners_x * cos_rot - self.corners_y * sin_rot
        y_rotated = self.corners_x * sin_rot + self.corners_y * cos_rot

        # Draw the square that we want to check if points are in.
        bb_path = mplPath.Path(
            np.array(
                [
                    [x_rotated[0], y_rotated[0]],
                    [x_rotated[1], y_rotated[1]],
                    [x_rotated[2], y_rotated[2]],
                    [x_rotated[3], y_rotated[3]],
                    [x_rotated[0], y_rotated[0]],
                ]
            ).astype(int)
        )

        ra_to_check, dec_to_check = _hpid2_ra_dec(self.nside, indices_to_check)

        # Project the indices to check to the tangent plane, see
        # if they fall inside the polygon
        x, y = gnomonic_project_toxy(ra_to_check, dec_to_check, ra, dec)
        x = (x * self.scale).astype(int)
        y = (y * self.scale).astype(int)
        for i, xcheck in enumerate(x):
            # I wonder if I can do this all at once rather than a loop?
            if bb_path.contains_point((x[i], y[i])):
                indices.append(indices_to_check[i])

        return np.array(indices)


def run_info_table(observatory, extra_info=None):
    """
    Make a little table for recording the information about a run
    """

    observatory_info = observatory.get_info()
    if extra_info is not None:
        for key in extra_info:
            observatory_info.append([key, extra_info[key]])
    observatory_info = np.array(observatory_info)

    n_feature_entries = 3

    names = ["Parameter", "Value"]
    dtypes = ["|U200", "|U200"]
    result = np.zeros(observatory_info[:, 0].size + n_feature_entries, dtype=list(zip(names, dtypes)))

    # Fill in info about the run
    result[0]["Parameter"] = "Date, ymd"
    now = datetime.datetime.now()
    result[0]["Value"] = "%i, %i, %i" % (now.year, now.month, now.day)

    result[1]["Parameter"] = "hostname"
    result[1]["Value"] = socket.gethostname()

    result[2]["Parameter"] = "rubin_scheduler.__version__"
    result[2]["Value"] = rsVersion.__version__

    result[3:]["Parameter"] = observatory_info[:, 0]
    result[3:]["Value"] = observatory_info[:, 1]

    return result


def inrange(inval, minimum=-1.0, maximum=1.0):
    """
    Make sure values are within min/max
    """
    inval = np.array(inval)
    below = np.where(inval < minimum)
    inval[below] = minimum
    above = np.where(inval > maximum)
    inval[above] = maximum
    return inval


def warm_start(scheduler, observations, mjd_key="mjd"):
    """Replay a list of observations into the scheduler

    Parameters
    ----------
    scheduler : scheduler object
    observations : np.array
        An array of observation (e.g., from sqlite2observations)
    """

    # Check that observations are in order
    observations.sort(order=mjd_key)
    for observation in observations:
        scheduler.add_observation(observation)

    return scheduler


def season_calc(night, offset=0, modulo=None, max_season=None, season_length=365.25, floor=True):
    """
    Compute what season a night is in with possible offset and modulo
    using convention that night -365 to 0 is season -1.

    Parameters
    ----------
    night : int or array
        The night we want to convert to a season
    offset : float or array (0)
        Offset to be applied to night (days)
    modulo : int (None)
        If the season should be modulated (i.e., so we can get all
        even years) (seasons, years w/default season_length)
    max_season : int (None)
        For any season above this value (before modulo), set to -1
    season_length : float (365.25)
        How long to consider one season (nights)
    floor : bool (True)
        If true, take the floor of the season. Otherwise, returns
        season as a float
    """
    if np.size(night) == 1:
        night = np.ravel(np.array([night]))
    result = night + offset
    result = result / season_length
    if floor:
        result = np.floor(result)
    if max_season is not None:
        over_indx = np.where(IntRounded(result) >= IntRounded(max_season))

    if modulo is not None:
        neg = np.where(IntRounded(result) < IntRounded(0))
        result = result % modulo
        result[neg] = -1
    if max_season is not None:
        result[over_indx] = -1
    if floor:
        result = result.astype(int)
    return result


def create_season_offset(nside, sun_ra_rad):
    """
    Make an offset map so seasons roll properly
    """
    hpindx = np.arange(hp.nside2npix(nside))
    ra, dec = _hpid2_ra_dec(nside, hpindx)
    offset = ra - sun_ra_rad + 2.0 * np.pi
    offset = offset % (np.pi * 2)
    offset = offset * 365.25 / (np.pi * 2)
    offset = -offset - 365.25
    return offset


class TargetoO:
    """Class to hold information about a target of opportunity object

    Parameters
    ----------
    tooid : int
        Unique ID for the ToO.
    footprints : np.array
        np.array healpix maps. 1 for areas to observe, 0 for no observe.
    mjd_start : float
        The MJD the ToO starts
    duration : float
        Duration of the ToO (days).
    ra_rad_center : float
        RA of the estimated center of the event (radians).
    dec_rad_center : float
        Dec of the estimated center of the event (radians).
    """

    def __init__(
        self,
        tooid,
        footprint,
        mjd_start,
        duration,
        ra_rad_center=None,
        dec_rad_center=None,
    ):
        self.footprint = footprint
        self.duration = duration
        self.id = tooid
        self.mjd_start = mjd_start
        self.ra_rad_center = ra_rad_center
        self.dec_rad_center = dec_rad_center


class SimTargetooServer:
    """Wrapper to deliver a targetoO object at the right time"""

    def __init__(self, targeto_o_list):
        self.targeto_o_list = targeto_o_list
        self.mjd_starts = np.array([too.mjd_start for too in self.targeto_o_list])
        durations = np.array([too.duration for too in self.targeto_o_list])
        self.mjd_ends = self.mjd_starts + durations

    def __call__(self, mjd):
        in_range = np.where((mjd > self.mjd_starts) & (mjd < self.mjd_ends))[0]
        result = None
        if in_range.size > 0:
            result = [self.targeto_o_list[i] for i in in_range]
        return result



__all__ = ("LongGapSurvey",)

import logging
from copy import copy, deepcopy

import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.surveys import BaseSurvey
from rubin_scheduler.scheduler.utils import scheduled_observation
from rubin_scheduler.utils import Site, _approx_ra_dec2_alt_az

log = logging.getLogger(__name__)


class LongGapSurvey(BaseSurvey):
    """
    Parameters
    ----------
    blob_survey : rubin_scheduler.scheduler.surveys.BlobSurvey
        A survey object that we will want to take repeat measurments of
        sometime later in the evening
    scripted_survey : rubin_scheduler.scheduler.surveys.ScriptedSurvey
        A scripted survey object that will have a queue updated with
        objects to observe later.
    gap range : list of 2 floats
        The desired gap range (hours)
    long_name : str
        The string to put in the observation 'note' for the scripted
        observations
    scripted_tol : float
        The tolerance for when scripted observations can execute (hours)
    after_meridian : bool (False)
        If True, force the scripted obsrevations to happen after they
        pass through the meridian. This can help make sure we don't hit
        the zenith exclusion zone.
    hour_step : float (0.5)
        The amount of time to step scheduled observations forward if
        they could try to execute in the zenith avoidance area (hours).
        Only used if `avoid_zenith` is True.
    ha_min(_max) : float (0,24)
        Trying to set so they don't acctually get used.

    """

    def __init__(
        self,
        blob_survey,
        scripted_survey,
        gap_range=[2, 10],
        long_name="long",
        scripted_tol=2.0,
        alt_min=20,
        alt_max=80.0,
        ha_min=24,
        ha_max=0.0,
        sun_alt_max=-18.0,
        flush_time=2.0,
        dist_tol=1.0,
        block_length=33.0,
        reverse=True,
        seed=42,
        night_max=50000,
        avoid_zenith=True,
        site=None,
        hour_step=0.5,
        survey_name=None,
    ):
        self.blob_survey = blob_survey
        self.scripted_survey = scripted_survey
        self.night = -1
        self.gap_range = np.array(gap_range) / 24.0  # To days
        rng = np.random.default_rng(seed)
        self.gaps = rng.uniform(self.gap_range.min(), self.gap_range.max(), night_max)
        self.gap = 0.0
        if site is None:
            self.site = Site("LSST")
        self.long_name = long_name
        if survey_name is None:
            self._generate_survey_name()
        else:
            self.survey_name = survey_name
        self.scripted_tol = scripted_tol / 24.0  # To days
        self.alt_min = np.radians(alt_min)
        self.alt_max = np.radians(alt_max)
        self.sun_alt_max = np.radians(sun_alt_max)
        self.ha_min = ha_min
        self.ha_max = ha_max
        self.flush_time = flush_time / 24.0
        self.dist_tol = np.radians(dist_tol)
        self.block_length = block_length / 60 / 24.0
        self.reverse = reverse
        self.avoid_zenith = avoid_zenith
        self.mjd_step = hour_step / 24.0

    def _generate_survey_name(self):
        self.survey_name = (
            f"Long Gap ({self.blob_survey.survey_name} +" f" {self.scripted_survey.survey_name})"
        )

    def _schedule_obs(self, observations):
        """Take incoming observations and decide if they should be added
        to the scripted survey to try and be observered again later
        """

        # Only match if we have completed the second of a pair and are
        # in most recent night.
        # ugh, stupid np.where doesn't support using scalars anymore
        if np.size(observations) == 1:
            if (observations["note"] == self.blob_survey.survey_note + ", b") & (
                observations["night"] == np.max(observations["night"])
            ):
                need_to_observe = np.array([0])
            else:
                need_to_observe = np.array([])
        else:
            need_to_observe = np.where(
                (observations["note"] == self.blob_survey.survey_note + ", b")
                & (observations["night"] == np.max(observations["night"]))
            )[0]

        # Set to the proper gap
        self.gap = self.gaps[np.max(observations["night"])]

        # If the incoming observation needs to have something
        # scheduled later
        if np.size(need_to_observe) > 0:
            sched_array = scheduled_observation(n=need_to_observe.size)
            for dt in np.intersect1d(observations.dtype.names, sched_array.dtype.names):
                if np.size(observations) == 1:
                    sched_array[dt] = observations[dt]
                else:
                    sched_array[dt] = observations[need_to_observe][dt]
            sched_array["mjd_tol"] = self.scripted_tol
            sched_array["alt_min"] = self.alt_min
            sched_array["alt_max"] = self.alt_max
            sched_array["HA_min"] = self.ha_min
            sched_array["HA_max"] = self.ha_max
            sched_array["sun_alt_max"] = self.sun_alt_max
            if np.size(observations) == 1:
                sched_array["flush_by_mjd"] = observations["mjd"] + self.flush_time + self.gap
                sched_array["mjd"] = observations["mjd"] + self.gap
            else:
                sched_array["flush_by_mjd"] = (
                    observations[need_to_observe]["mjd"] + self.flush_time + self.gap
                )
                sched_array["mjd"] = observations[need_to_observe]["mjd"] + self.gap
            sched_array["dist_tol"] = self.dist_tol
            if self.avoid_zenith:
                # when is the earliest we expect things could execute
                earliest_mjd = sched_array["mjd"] - sched_array["mjd_tol"]
                alts = []
                mjds = np.arange(
                    np.min(earliest_mjd),
                    np.max(sched_array["mjd"]) + self.mjd_step,
                    self.mjd_step,
                )
                # Let's compute the alt of everything at earliest and
                # scheduled
                for mjd in mjds:
                    alt, az = _approx_ra_dec2_alt_az(
                        sched_array["RA"],
                        sched_array["dec"],
                        self.site.latitude_rad,
                        self.site.longitude_rad,
                        mjd,
                    )
                    alts.append(np.max(alt))

                while np.max(alts) > self.alt_max:
                    alts = []
                    sched_array["mjd"] += self.mjd_step
                    sched_array["flush_by_mjd"] += self.mjd_step
                    earliest_mjd = sched_array["mjd"] - sched_array["mjd_tol"]
                    # Let's compute the alt of everything then
                    mjds = np.arange(
                        np.min(earliest_mjd),
                        np.max(sched_array["mjd"]) + self.mjd_step,
                        self.mjd_step,
                    )
                    # Let's compute the alt of everything at earliest
                    # and scheduled
                    for mjd in mjds:
                        alt, az = _approx_ra_dec2_alt_az(
                            sched_array["RA"],
                            sched_array["dec"],
                            self.site.latitude_rad,
                            self.site.longitude_rad,
                            mjd,
                        )
                        alts.append(np.max(alt))
            # Make sure these have the note filled in
            sched_array["note"] = self.long_name

            # See if we need to append things to the scripted survey
            # object
            if self.scripted_survey.obs_wanted is not None:
                sched_array = np.concatenate([self.scripted_survey.obs_wanted, sched_array])

            self.scripted_survey.set_script(sched_array)

    def add_observations_array(self, observations_array_in, observations_hpid_in):
        self._schedule_obs(observations_array_in)

        self.blob_survey.add_observations_array(observations_array_in, observations_hpid_in)
        self.scripted_survey.add_observations_array(observations_array_in, observations_hpid_in)

    def add_observation(self, observation, **kwargs):
        self._schedule_obs(observation)

        self.blob_survey.add_observation(observation, **kwargs)
        self.scripted_survey.add_observation(observation, **kwargs)

    def _check_feasibility(self, conditions):
        f1 = self.blob_survey._check_feasibility(conditions)
        f2 = self.scripted_survey._check_feasibility(conditions)

        # If either one is able to go, we can observe
        result = f1 | f2
        return result

    def get_scheduled_obs(self):
        return self.scripted_survey.get_scheduled_obs()

    def calc_reward_function(self, conditions):
        if conditions.night != self.night:
            # Clear out the scheduled observations
            self.scripted_survey.clear_script()
            self.night = copy(conditions.night)
            self.gap = self.gaps[conditions.night]
            time_remaining = conditions.sun_n18_rising - conditions.mjd
            if self.gap > time_remaining:
                self.gap = time_remaining - self.block_length
            # XXX-need to reach into the blob and set what the gap
            # is I guess

        self.r1 = self.blob_survey.calc_reward_function(conditions)
        self.r2 = self.scripted_survey.calc_reward_function(conditions)
        self.reward_checked = True
        return np.nanmax([np.nanmax(self.r1), np.nanmax(self.r2)])

    def generate_observations_rough(self, conditions):
        """"""
        pass

    def generate_observations(self, conditions):
        if not self.reward_checked:
            self.reward = self.calc_reward_function(conditions)

        # Check for any pre-scheduled
        if self.r2 > -np.inf:
            observations = self.scripted_survey.generate_observations(conditions)
        else:
            observations = None

        # Check if it's a good time for a blob
        if observations is None:
            observations = self.blob_survey.generate_observations(conditions)

        return observations

    def make_reward_df(self, conditions, accum=True):
        """Create a pandas.DataFrame describing the reward from the survey.

        Parameters
        ----------
        conditions : `rubin_scheduler.scheduler.features.Conditions`
            Conditions for which rewards are to be returned
        accum : `bool`
            Include accumulated rewards

        Returns
        -------
        reward_df : `pandas.DataFrame`
            A table of surveys listing the rewards.
        """

        test_survey = deepcopy(self)
        reward = test_survey.calc_reward_function(conditions)
        feasible = test_survey._check_feasibility(conditions) and reward > np.finfo(reward).min

        if accum:
            reward_df = pd.DataFrame(
                {
                    "basis_function": ["None"],
                    "feasible": [feasible],
                    "max_basis_reward": [reward],
                    "basis_area": [0],
                    "basis_weight": [1],
                    "max_accum_reward": [reward],
                    "accum_area": [0],
                }
            )
        else:
            reward_df = pd.DataFrame(
                {
                    "basis_function": ["None"],
                    "feasible": [feasible],
                    "max_basis_reward": [reward],
                    "basis_area": [0],
                    "basis_weight": [1],
                }
            )

        return reward_df



"""
This file contains coordinate transformation methods and utilities
for converting an ra,dec coordinate set to cartesian coordinates and
to grid id using a spatial tree.
"""

__all__ = ("_build_tree",)

import numpy as np
from scipy.spatial import cKDTree as kdTree

from rubin_scheduler.utils.coordinate_transformations import _xyz_from_ra_dec


def _build_tree(ra, dec, leafsize=100, scale=None):
    """
    Build KD tree on simDataRA/Dec and set radius (via set_rad) for
    matching.

    Parameters
    ----------
    ra, dec : float (or arrays)
        RA and Dec values (in radians).
    leafsize : int (100)
        The number of Ra/Dec pointings in each leaf node.
    scale : float (None)
        If set, the values are scaled up, rounded, and converted to
        integers. Useful for forcing a set precision and preventing
        machine precision differences
    """
    if np.any(np.abs(ra) > np.pi * 2.0) or np.any(np.abs(dec) > np.pi * 2.0):
        raise ValueError("Expecting RA and Dec values to be in radians.")
    x, y, z = _xyz_from_ra_dec(ra, dec)
    if scale is not None:
        x = np.round(x * scale).astype(int)
        y = np.round(y * scale).astype(int)
        z = np.round(z * scale).astype(int)
    data = list(zip(x, y, z))
    if np.size(data) > 0:
        try:
            tree = kdTree(data, leafsize=leafsize, balanced_tree=False, compact_nodes=False)
        except TypeError:
            tree = kdTree(data, leafsize=leafsize)
    else:
        raise ValueError("ra and dec should have length greater than 0.")

    return tree



"""
This is a dummy module that creates global caches so that we can test
the behavior of sims_clean_up()
"""

__all__ = ("a_dict_cache", "a_list_cache")

from rubin_scheduler.utils.code_utilities import sims_clean_up

a_dict_cache = {}

a_list_cache = []

sims_clean_up.targets.append(a_dict_cache)
sims_clean_up.targets.append(a_list_cache)



import unittest

import healpy as hp
import numpy as np
import pandas as pd

import rubin_scheduler.scheduler.basis_functions as basis_functions
import rubin_scheduler.scheduler.surveys as surveys
from rubin_scheduler.scheduler.basis_functions import SimpleArrayBasisFunction
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.utils import empty_observation, set_default_nside


class TestSurveys(unittest.TestCase):
    def test_field_survey(self):
        nside = 32

        bfs = []
        bfs.append(basis_functions.M5DiffBasisFunction(nside=nside))
        survey = surveys.FieldSurvey(bfs, RA=90.0, dec=-30.0, reward_value=1)

        observatory = ModelObservatory()

        # Check dunder methods
        self.assertIsInstance(repr(survey), str)
        self.assertIsInstance(str(survey), str)

        # Check survey access methods
        conditions = observatory.return_conditions()
        reward = survey.calc_reward_function(conditions)
        self.assertIsInstance(reward, float)
        reward_df = survey.reward_changes(conditions)
        reward_df = survey.make_reward_df(conditions)
        self.assertIsInstance(reward_df, pd.DataFrame)
        reward_df = survey.make_reward_df(conditions, accum=False)

    def test_pointings_survey(self):
        """Test the pointing survey."""
        mo = ModelObservatory()
        conditions = mo.return_conditions()

        # Make a ring of points near the equator so
        # some should always be visible
        fields = empty_observation(n=10)
        fields["RA"] = np.arange(0, fields.size) / fields.size * 2.0 * np.pi
        fields["dec"] = -0.01
        fields["note"] = ["test%i" % ind for ind in range(fields.size)]
        fields["filter"] = "r"
        survey = surveys.PointingsSurvey(fields)

        reward = survey.calc_reward_function(conditions)
        assert np.isfinite(reward)

        obs = survey.generate_observations(conditions)
        # Confirm that our desired input values got passed through
        assert obs[0]["dec"] < 0
        assert obs[0]["note"][0][0:4] == "test"

        # Adding observations
        assert np.sum(survey.n_obs) == 0
        survey.add_observation(obs[0])
        assert np.sum(survey.n_obs) == 1
        survey.add_observations_array(fields, None)
        assert np.sum(survey.n_obs) == 11

        # Check we can get display things out
        rc = survey.reward_changes(conditions)
        assert len(rc) == len(survey.weights)

        # Check we get a dataFrame
        df = survey.make_reward_df(conditions)
        assert len(df) == len(survey.weights)

    def test_roi(self):
        random_seed = 6563
        infeasible_hpix = 123
        nside = set_default_nside()
        npix = hp.nside2npix(nside)
        rng = np.random.default_rng(seed=random_seed)
        num_bfs = 3
        bf_values = rng.random((num_bfs, npix))
        bf_values[:, infeasible_hpix] = -np.inf
        bfs = [SimpleArrayBasisFunction(values) for values in bf_values]

        observatory = ModelObservatory()
        conditions = observatory.return_conditions()

        # A few cases with an ROI with one valid healpix
        for i in range(3):
            hpix = rng.integers(npix)
            ra, decl = hp.pix2ang(nside, hpix, lonlat=True)
            survey = surveys.FieldSurvey(bfs, RA=ra, dec=decl, reward_value=1)
            reward_df = survey.make_reward_df(conditions)
            for value, max_basis_reward in zip(bf_values[:, hpix], reward_df["max_basis_reward"]):
                self.assertEqual(max_basis_reward, value)

        # One case with an ROI with only an infeasible healpix
        ra, decl = hp.pix2ang(nside, infeasible_hpix, lonlat=True)
        survey = surveys.FieldSurvey(bfs, RA=ra, dec=decl, reward_value=1)
        reward_df = survey.make_reward_df(conditions)
        for max_basis_reward in reward_df["max_basis_reward"]:
            self.assertEqual(max_basis_reward, -np.inf)

        for area in reward_df["basis_area"]:
            self.assertEqual(area, 0.0)

        for feasible in reward_df["feasible"]:
            self.assertFalse(feasible)

        # Make sure it still works as expected if no ROI is set
        weights = [1] * num_bfs
        survey = surveys.BaseMarkovSurvey(bfs, weights)
        reward_df = survey.make_reward_df(conditions)
        for value, max_basis_reward in zip(bf_values.max(axis=1), reward_df["max_basis_reward"]):
            self.assertEqual(max_basis_reward, value)


if __name__ == "__main__":
    unittest.main()



__all__ = ("SkyModelPre", "interp_angle")

import abc
import glob
import os
import urllib
import warnings
from pathlib import Path

import h5py
import healpy as hp
import numpy as np
import requests
from astropy import units as u
from astropy.coordinates import AltAz, EarthLocation, SkyCoord, get_sun
from astropy.time import Time

import rubin_scheduler.data.rs_download_sky
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import Site, _angular_separation, _hpid2_ra_dec, survey_start_mjd

try:
    from lsst.resources import ResourcePath
except ImportError:
    pass


def short_angle_dist(a0, a1):
    """
    from https://gist.github.com/shaunlebron/8832585
    """
    max_angle = 2.0 * np.pi
    da = (a1 - a0) % max_angle
    return 2.0 * da % max_angle - da


def interp_angle(x_out, xp, anglep, degrees=False):
    """
    Interpolate angle values (handle wrap around properly).
    Does nearest neighbor interpolation if values out of range.

    Parameters
    ----------
    x_out : `float` or array
        The points to interpolate to.
    xp : array
        Points to interpolate between (must be sorted)
    anglep : array
        The angles ascociated with xp
    degrees : `bool` (False)
        Set if anglep is degrees (True) or radidian (False)
    """

    # Where are the interpolation points
    x = np.atleast_1d(x_out)
    left = np.searchsorted(xp, x) - 1
    right = left + 1

    # If we are out of bounds, just use the edges
    right[np.where(right >= xp.size)] -= 1
    left[np.where(left < 0)] += 1
    baseline = xp[right] - xp[left]

    wterm = (x - xp[left]) / baseline
    wterm[np.where(baseline == 0)] = 0
    if degrees:
        result = (
            np.radians(anglep[left])
            + short_angle_dist(np.radians(anglep[left]), np.radians(anglep[right])) * wterm
        )
        result = result % (2.0 * np.pi)
        result = np.degrees(result)
    else:
        result = anglep[left] + short_angle_dist(anglep[left], anglep[right]) * wterm
        result = result % (2.0 * np.pi)
    return result


def simple_daytime(sky_alt, sky_az, sun_alt, sun_az, filter_name="r", bright_val=2.0, sky_alt_min=20.0):
    """A simple function to return a sky brightness map when the sun is up

    Parameters
    ----------
    sky_alt : `float`
        Altitude of poistion(s) on the sky. Degrees.
    sky_az : `float`
        Azimuth of poistion(s) on the sky. Degrees.
    sun_alt : `float`
        Altitude of the sun. Degrees.
    sun_az : `float`
        Azimuth of the sun. Degrees.
    filter_name : `str`
        Name of the filter, default "r". Currently unused, but
        could be useful in the future if a more complicated function
        gets subbed in.
    bright_val : `float`
        The value to plug into the sky. Default 2 mag/sq arcsec.
    sky_alt_min : `float`
        Set all sky below the alt limit to NaN. Default 20 degrees.
    """

    result = np.full_like(sky_alt, np.nan)
    result[np.where(sky_alt > sky_alt_min)] = bright_val
    return result


class SkyModelPreBase(abc.ABC):
    """Load pre-computed sky brighntess maps for the LSST site
    and use them to interpolate to arbitrary dates.

    Parameters
    ----------
    data_path : `str`, opt
        path to the numpy save files. Looks in standard
        SIMS_SKYBRIGHTNESS_DATA or RUBIN_SIM_DATA_DIR if set to
        default (None).
    init_load_length : `int` (10)
        The length of time (days) to load from disk initially.
        Set to something small for fast reads.
    load_length : `int` (365)
        The number of days to load after the initial load.
    mjd0 : `float` (None)
        The starting MJD to load on initilization (days). Uses
        util to lookup default if None.
    location : `astropy.EarthLocation`
        The location of the telescope. Default of None will load
        Rubin position.
    sun_alt_limit : `float`
            The altitude limit to use a "bright" sky function. Default
            of -8 degrees
    """

    def __init__(
        self,
        data_path=None,
        init_load_length=10,
        load_length=365,
        verbose=False,
        mjd0=None,
        location=None,
        sun_alt_limit=-8.0,
    ):
        self.info = None
        self.sb = None
        self.header = None
        self.filter_names = None
        self.verbose = verbose
        self.sun_alt_limit = np.radians(sun_alt_limit)

        if location is None:
            site = Site("LSST")
            self.location = EarthLocation(lat=site.latitude, lon=site.longitude, height=site.height)
        else:
            self.location = location

        # Look in default location for .npz files to load
        if data_path is not None:
            self.data_path = data_path
        elif "SIMS_SKYBRIGHTNESS_DATA" in os.environ:
            self.data_path = os.environ["SIMS_SKYBRIGHTNESS_DATA"]
        else:
            self.data_path = os.path.join(get_data_dir(), "skybrightness_pre")

        self._init_files()

        if len(self.files) == 0:
            errmssg = "Failed to find pre-computed .h5 files. "
            errmssg += "Copy data from NCSA with sims_skybrightness_pre/data/data_down.sh \n"
            errmssg += "or build by running sims_skybrightness_pre/data/generate_hdf5.py"
            warnings.warn(errmssg)

        self._init_filesizes()
        self._init_file_mjd_ranges()

        # Set that nothing is loaded at this point
        self.loaded_range = np.array([-1])
        self.timestep_max = -1

        if mjd0 is None:
            mjd0 = survey_start_mjd()

        # Do a quick initial load if set
        if init_load_length is not None:
            self.load_length = init_load_length
            self._load_data(mjd0)
        # swap back to the longer load length
        self.load_length = load_length
        self.nside = 32
        hpid = np.arange(hp.nside2npix(self.nside))
        self.ra, self.dec = _hpid2_ra_dec(self.nside, hpid)
        self.skycoord = SkyCoord(ra=self.ra * u.rad, dec=self.dec * u.rad)

    def _load_data(self, mjd, filename=None, npyfile=None):
        """Load up the h5 file to interpolate things.

        Parameters
        ----------
        mjd : `float`
            The Modified Julian Date that we want to load
        filename : `str` (None)
            The filename to restore. If None, it checks the filenames
            on disk to find one that should have the requested MJD
        npyfile : `str` (None)
            If sky brightness data not in npz file, checks the .npy
            file with same root name.
        """

        if filename is None:
            # Figure out which file to load.
            file_indx = np.where((mjd >= self.mjd_left) & (mjd <= self.mjd_right))[0]
            if np.size(file_indx) == 0:
                raise ValueError(
                    "MJD = %f is out of range for the files found (%f-%f)"
                    % (mjd, self.mjd_left.min(), self.mjd_right.max())
                )
            # Just take the later one, assuming we're probably
            # simulating forward in time
            file_indx = np.max(file_indx)

            filename = self.files[file_indx]
        else:
            self.loaded_range = None

        # Use three separate try/excepet blocks so that if any of
        # them throw exceptions, we still get the others.
        try:
            del self.sb
        except AttributeError:
            pass

        try:
            del self.filter_names
        except AttributeError:
            pass

        try:
            del self.timestep_max
        except AttributeError:
            pass

        if self.verbose:
            print("Loading file %s" % filename)
        h5 = self._create_h5(filename, "r")
        mjds = h5["mjds"][:]
        indxs = np.where((mjds >= mjd) & (mjds <= (mjd + self.load_length)))
        indxs = [np.min(indxs), np.max(indxs)]
        if indxs[0] > 0:
            indxs[0] -= 1
        self.loaded_range = np.array([mjds[indxs[0]], mjds[indxs[1]]])
        self.mjds = mjds[indxs[0] : indxs[1]]
        _timestep_max = np.empty(1, dtype=float)
        h5["timestep_max"].read_direct(_timestep_max)
        self.timestep_max = np.max(_timestep_max)

        self.sb = h5["sky_mags"][indxs[0] : indxs[1]]
        self.filter_names = self.sb.dtype.names
        h5.close()

        if self.verbose:
            print("%s loaded" % os.path.split(filename)[1])

        self.nside = hp.npix2nside(self.sb[self.filter_names[0]][0, :].size)

    def return_mags(
        self,
        mjd,
        indx=None,
        badval=hp.UNSEEN,
        filters=["u", "g", "r", "i", "z", "y"],
        extrapolate=False,
    ):
        """Return a full sky map or individual pixels for the input mjd.

        Parameters
        ----------
        mjd : `float`
            Modified Julian Date to interpolate to
        indx : `List` of `int` (None)
            indices to interpolate the sky values at. Returns full sky
            if None. If the class was instatiated with opsimFields,
            indx is the field ID, otherwise it is the healpix ID.
        airmass_mask : `bool` (True)
            Set high (>2.5) airmass pixels to badval.
        planet_mask : `bool` (True)
            Set sky maps to badval near (2 degrees) bright planets.
        moon_mask : `bool` (True)
            Set sky maps near (10 degrees) the moon to badval.
        zenith_mask : `bool` (True)
            Set sky maps at high altitude (>86.5) to badval.
        badval : `float` (-1.6375e30)
            Mask value. Defaults to the healpy mask value.
        filters : `list`, opt
            List of strings for the filters that should be returned.
            Default returns ugrizy.
        extrapolate : `bool` (False)
            In indx is set, extrapolate any masked pixels to be the
            same as the nearest non-masked value from the full sky map.

        Returns
        -------
        sbs : `dict`
            A dictionary with filter names as keys and np.arrays as
            values which hold the sky brightness maps in mag/sq arcsec.
        """
        if mjd < self.loaded_range.min() or (mjd > self.loaded_range.max()):
            self._load_data(mjd)

        left = np.searchsorted(self.mjds, mjd) - 1
        right = left + 1

        # Do full sky by default
        if indx is None:
            indx = np.arange(self.sb["r"].shape[1])
            full_sky = True
        else:
            full_sky = False

        # If we are out of bounds
        if right >= self.mjds.size:
            right -= 1
            baseline = 1.0
        elif left < 0:
            left += 1
            baseline = 1.0
        else:
            baseline = self.mjds[right] - self.mjds[left]

        # Check if we are between sunrise/set
        if baseline > self.timestep_max + 1e-6:

            # Check if sun is really high:
            obstime = Time(mjd, format="mjd")
            sun = get_sun(obstime)
            aa = AltAz(location=self.location, obstime=obstime)
            sun_alt_az = sun.transform_to(aa)

            if sun_alt_az.alt.rad > self.sun_alt_limit:
                warnings.warn("Sun high, using bright sky approx")
                hp_aa = self.skycoord.transform_to(aa)

                sbs = {}
                for filter_name in filters:
                    sbs[filter_name] = simple_daytime(
                        hp_aa.alt.deg,
                        hp_aa.az.deg,
                        sun_alt_az.alt.deg,
                        sun_alt_az.az.deg,
                        filter_name=filter_name,
                    )

            else:
                warnings.warn("Requested MJD between sunrise and sunset, returning closest maps")
                diff = np.abs(self.mjds[left.max() : right.max() + 1] - mjd)
                closest_indx = np.array([left, right])[np.where(diff == np.min(diff))].min()
                sbs = {}
                for filter_name in filters:
                    sbs[filter_name] = self.sb[filter_name][closest_indx, indx]
                    sbs[filter_name][np.isinf(sbs[filter_name])] = badval
                    sbs[filter_name][np.where(sbs[filter_name] == hp.UNSEEN)] = badval
        else:
            wterm = (mjd - self.mjds[left]) / baseline
            w1 = 1.0 - wterm
            w2 = wterm
            sbs = {}
            for filter_name in filters:
                sbs[filter_name] = (
                    self.sb[filter_name][left, indx] * w1 + self.sb[filter_name][right, indx] * w2
                )
        # If requested a certain pixel(s), and want to extrapolate.
        if (not full_sky) & extrapolate:
            masked_pix = False
            for filter_name in filters:
                if (badval in sbs[filter_name]) | (True in np.isnan(sbs[filter_name])):
                    masked_pix = True
            if masked_pix:
                # We have pixels that are masked that we want
                # reasonable values for
                full_sky_sb = self.return_mags(
                    mjd,
                    filters=filters,
                )
                good = np.where((full_sky_sb[filters[0]] != badval) & ~np.isnan(full_sky_sb[filters[0]]))[0]
                ra_full = self.ra[good]
                dec_full = self.dec[good]
                for filtername in filters:
                    full_sky_sb[filtername] = full_sky_sb[filtername][good]
                # Going to assume the masked pixels are the same in all filters
                masked_indx = np.where(
                    (sbs[filters[0]].ravel() == badval) | np.isnan(sbs[filters[0]].ravel())
                )[0]
                for i, mi in enumerate(masked_indx):
                    # Note, this is going to be really slow for many
                    # pixels, should use a kdtree
                    dist = _angular_separation(
                        self.ra[indx[i]],
                        self.dec[indx[i]],
                        ra_full,
                        dec_full,
                    )
                    closest = np.where(dist == dist.min())[0]
                    for filtername in filters:
                        sbs[filtername].ravel()[mi] = np.min(full_sky_sb[filtername][closest])

        return sbs

    @abc.abstractmethod
    def _init_files(self):
        pass

    @abc.abstractmethod
    def _init_filesizes(self):
        pass

    @abc.abstractmethod
    def _init_file_mjd_ranges(self):
        pass

    @abc.abstractmethod
    def _create_h5(self, filename, *args, **kwargs):
        pass


class SkyModelPreWithLocalFilesOnly(SkyModelPreBase):
    def _init_files(self):
        self.files = glob.glob(os.path.join(self.data_path, "*.h5"))
        self.files.sort()

    def _init_filesizes(self):
        self.filesizes = np.array([os.path.getsize(filename) for filename in self.files])

    def _init_file_mjd_ranges(self):
        mjd_left = []
        mjd_right = []
        for filename in self.files:
            temp = os.path.split(filename)[-1].replace(".h5", "").split("_")
            mjd_left.append(float(temp[0]))
            mjd_right.append(float(temp[1]))

        self.mjd_left = np.array(mjd_left)
        self.mjd_right = np.array(mjd_right)

    def _create_h5(self, filename, *args, **kwargs):
        # This method exists so it can be overriden in a subclass
        h5 = h5py.File(filename, *args, **kwargs)
        return h5


class SkyModelPreWithResources(SkyModelPreBase):
    def _init_files(self):
        data_resource = ResourcePath(self.data_path, forceDirectory=True)

        try:
            self.files = list(ResourcePath.findFileResources([data_resource], file_filter=r".*\.h5"))
        except NotImplementedError:
            # lsst.requests does not implement walk for plain html,
            # so do it manually here.
            # Unlike the method above, however, this simple approach
            # does not descend into subdirectories.
            self.files = []
            if urllib.parse.urlparse(data_resource.geturl()).scheme in ("http", "https"):
                request_content = requests.get(data_resource.geturl()).text
                html_parser = rubin_scheduler.data.rs_download_sky.MyHTMLParser()
                html_parser.feed(request_content)
                html_parser.close()
                for file_name in html_parser.filenames:
                    if file_name.endswith(".h5"):
                        self.files.append(data_resource.join(file_name))
            else:
                raise

        self.files.sort()

    def _init_filesizes(self):
        self.filesizes = np.array([file_path.size() for file_path in self.files])

    def _init_file_mjd_ranges(self):
        mjd_left = []
        mjd_right = []
        for file_resource_path in self.files:
            temp = Path(file_resource_path.split()[1]).stem.split("_")
            mjd_left.append(float(temp[0]))
            mjd_right.append(float(temp[1]))

        self.mjd_left = np.array(mjd_left)
        self.mjd_right = np.array(mjd_right)

    def _create_h5(self, filename, *args, **kwargs):
        with filename.as_local() as local_file_resource_path:
            h5 = h5py.File(local_file_resource_path.ospath, *args, **kwargs)
        return h5


if "ResourcePath" in dir():

    class SkyModelPre(SkyModelPreWithResources):
        pass

else:

    class SkyModelPre(SkyModelPreWithLocalFilesOnly):
        pass



__all__ = ("PlanAheadSurvey",)

from copy import copy

import healpy as hp
import numpy as np

import rubin_scheduler.scheduler.basis_functions as bfs
from rubin_scheduler.scheduler import features

from .surveys import BlobSurvey


class PlanAheadSurvey(BlobSurvey):
    """Have a survey object that can plan ahead if it will want to
    observer a blob later in the night

    Parameters
    ----------
    delta_mjd_tol : float
        The tolerance to alow on when to execute scheduled observations
        (days)
    minimum_sky_area : float
        The minimum sky area to demand before making a scheduled
        observation (square degrees)
    track_filters : str
        The filter name we want to prevent long gaps on
    in_season : float
        The distance in RA from the meridian at midnight to consider
        (hours). This is the half-width
    cadence : float
        Ignore gaps below this length (days)
    """

    def __init__(
        self,
        basis_functions,
        basis_weights,
        delta_mjd_tol=0.3 / 24.0,
        minimum_sky_area=200.0,
        track_filters="g",
        in_season=2.5,
        cadence=9,
        **kwargs,
    ):
        super(PlanAheadSurvey, self).__init__(basis_functions, basis_weights, **kwargs)
        # note that self.night is already being used for tracking
        # tesselation. So here's an attribute for seeing if the night
        # has changed for cadence tracking
        self.night_cad = -100
        self.scheduled_obs = None
        self.delta_mjd_tol = delta_mjd_tol
        self.minimum_sky_area = minimum_sky_area  # sq degrees
        self.extra_features = {}
        self.extra_features["last_observed"] = features.Last_observed(filtername=track_filters)
        self.extra_basis_functions = {}
        self.extra_basis_functions["moon_mask"] = bfs.Moon_avoidance_basis_function()
        self.track_filters = track_filters
        self.in_season = in_season / 12.0 * np.pi  # to radians

        self.pix_area = hp.nside2pixarea(self.nside, degrees=True)
        self.cadence = cadence

    def check_night(self, conditions):
        """"""
        delta_mjd = conditions.mjd - self.extra_features["last_observed"].feature
        moon_mask = self.extra_basis_functions["moon_mask"](conditions)

        pix_to_obs = np.where(
            (delta_mjd > self.cadence)
            & (np.abs(conditions.az_to_antisun) < self.in_season)
            & (moon_mask >= 0)
        )[0]

        area = np.size(pix_to_obs) * self.pix_area

        # If there are going to be some observations at a given time
        if area > self.minimum_sky_area:
            # Maybe just calculate the mean (with angles)
            # Via https://en.wikipedia.org/wiki/
            # Mean_of_circular_quantities
            mean_ra = np.arctan2(
                np.mean(np.sin(conditions.ra[pix_to_obs])),
                np.mean(np.cos(conditions.ra[pix_to_obs])),
            )
            if mean_ra < 0:
                mean_ra += 2.0 * np.pi

            hour_angle = conditions.lmst - mean_ra * 12.0 / np.pi
            # This should be running -12 hours to +12 hours

            hour_angle[np.where(hour_angle < -12)] += 24
            hour_angle[np.where(hour_angle > 12)] -= 24

            if hour_angle < 0:
                self.scheduled_obs = conditions.mjd - hour_angle / 24.0
            else:
                self.scheduled_obs = conditions.mjd
        else:
            self.scheduled_obs = None

    def calc_reward_function(self, conditions):
        # Only compute if we will want to observe sometime in the night
        self.reward = -np.inf
        if self._check_feasibility(conditions):
            if self.night_cad != conditions.night:
                self.check_night(conditions)
                self.night_cad = copy(conditions.night)

            if self.scheduled_obs is not None:
                # If there are scheduled observations, and we are in
                # the correct time window
                delta_mjd = conditions.mjd - self.scheduled_obs
                if (np.abs(delta_mjd) < self.delta_mjd_tol) & (self.scheduled_obs is not None):
                    # So, we think there's a region that has had a
                    # long gap and can be observed
                    # call the standard reward function
                    self.reward = super(PlanAheadSurvey, self).calc_reward_function(conditions)

        return self.reward

    def generate_observations(self, conditions):
        observations = super(PlanAheadSurvey, self).generate_observations(conditions)
        # We are providing observations, so clear the scheduled obs
        # and reset the night so
        # self.check_night will get called again in case there's
        # another blob that should be done after this one completes
        self.scheduled_obs = None
        self.night_cad = conditions.night - 1
        return observations



import os
import unittest

from astropy.time import Time, TimeDelta

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.site_models import CloudData


class TestCloudModel(unittest.TestCase):
    def setUp(self):
        self.th = Time("2020-01-01", format="isot", scale="tai")
        self.cloud_db = os.path.join(get_data_dir(), "site_models", "clouds_ctio_1975_2022.db")
        self.num_original_values = 69653

    def test_basic_information_after_creation(self):
        cloud_data = CloudData(self.th, cloud_db=self.cloud_db)
        self.assertEqual(cloud_data.start_time, self.th)
        cloud_data = CloudData(self.th, cloud_db=self.cloud_db, offset_year=1)
        self.assertEqual(cloud_data.start_time, Time("2021-01-01", format="isot", scale="tai"))

    def test_information_after_initialization(self):
        # Test setting cloud_db explicitly.
        cloud_data = CloudData(self.th, cloud_db=self.cloud_db)
        cloud_data.read_data()
        self.assertEqual(cloud_data.cloud_values.size, self.num_original_values)
        self.assertEqual(cloud_data.cloud_dates.size, self.num_original_values)
        # Test that find built-in module automatically.
        cloud_data = CloudData(self.th)
        cloud_data.read_data()
        self.assertEqual(cloud_data.cloud_dates.size, self.num_original_values)

    def test_get_clouds(self):
        cloud_data = CloudData(self.th, cloud_db=self.cloud_db)
        cloud_data.read_data()
        dt = TimeDelta(700000, format="sec")
        self.assertEqual(cloud_data(self.th + dt), 0.0)
        dt = TimeDelta(701500, format="sec")
        self.assertEqual(cloud_data(self.th + dt), 0.0)
        dt = TimeDelta(705000, format="sec")
        self.assertEqual(cloud_data(self.th + dt), 0.0)
        dt = TimeDelta(6306840, format="sec")
        self.assertEqual(cloud_data(self.th + dt), 0.0)

    def test_get_clouds_using_different_start_month(self):
        # Just changing the starting month
        t2 = Time("2020-05-24", format="isot", scale="tai")
        cloud1 = CloudData(t2, cloud_db=self.cloud_db)
        self.assertEqual(cloud1.start_time, self.th)
        cloud1.read_data()
        dt = TimeDelta(700000, format="sec")
        self.assertEqual(cloud1(t2 + dt), 0.25)
        dt = TimeDelta(6306840, format="sec")
        self.assertEqual(cloud1(t2 + dt), 1.0)


if __name__ == "__main__":
    unittest.main()



__all__ = ("data_dict", "scheduler_download_data", "download_rubin_data", "DEFAULT_DATA_URL")

import argparse
import os
import warnings
from shutil import rmtree, unpack_archive

import requests
from requests.exceptions import ConnectionError
from tqdm.auto import tqdm

from .data_sets import data_versions, get_data_dir

DEFAULT_DATA_URL = "https://s3df.slac.stanford.edu/data/rubin/sim-data/rubin_sim_data/"


def data_dict():
    """Creates a `dict` for all data buckets and the tar file they map to.
    To create tar files and follow any sym links, run:
    `tar -chvzf maf_may_2021.tgz maf`

    Returns
    -------
    result : `dict`
        Data bucket filenames dictionary with keys/values:
        "name" - Data bucket name (`str`).
        "version" - Versioned file name (`str`).
    """
    file_dict = {
        "scheduler": "scheduler_2023_10_16.tgz",
        "site_models": "site_models_2023_10_02.tgz",
        "skybrightness_pre": "skybrightness_pre_2023_10_17.tgz",
        "utils": "utils_2023_11_02.tgz",
    }
    return file_dict


def scheduler_download_data(file_dict=None):
    """Download data."""

    if file_dict is None:
        file_dict = data_dict()
    parser = argparse.ArgumentParser(description="Download data files for rubin_sim package")
    parser.add_argument(
        "--versions",
        dest="versions",
        default=False,
        action="store_true",
        help="Report expected versions, then quit",
    )
    parser.add_argument(
        "--update",
        dest="update",
        default=False,
        action="store_true",
        help="Update versions of data on disk to match current",
    )
    parser.add_argument(
        "-d",
        "--dirs",
        type=str,
        default=None,
        help="Comma-separated list of directories to download",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        default=False,
        action="store_true",
        help="Force re-download of data directory(ies)",
    )
    parser.add_argument(
        "--url_base",
        type=str,
        default=DEFAULT_DATA_URL,
        help="Root URL of download location",
    )
    parser.add_argument(
        "--tdqm_disable",
        dest="tdqm_disable",
        default=False,
        action="store_true",
        help="Turn off tdqm progress bar",
    )
    args = parser.parse_args()

    download_rubin_data(
        data_dict(),
        dirs=args.dirs,
        print_versions_only=args.versions,
        update=args.update,
        force=args.force,
        url_base=args.url_base,
        tdqm_disable=args.tdqm_disable,
    )


def download_rubin_data(
    file_dict,
    dirs=None,
    print_versions_only=False,
    update=False,
    force=False,
    url_base=DEFAULT_DATA_URL,
    tdqm_disable=False,
):
    """Download external data blobs

    Parameters
    ----------
    file_dict : `dict`
        A dict with keys of directory names and values of remote filenames.
    dirs : `list` [`str`]
        List of directories to download. Default (None) assumes they are
        in file_dict
    versions : `bool`
        If True, print the versions currently on disk. Default False.
    update : `bool`
        If True, update versions on disk to match expected 'current'.
        Default False.
    force : `bool`
        If True, replace versions on disk with new download. Default False.
    url_base : `str`
        The URL to use, default to DEFAULT_DATA_URL
    tdqm_disable : `bool`
        If True, disable the tdqm progress bar. Default False.
    """
    # file_dict = dictionary of current versions
    if dirs is None:
        dirs = file_dict.keys()
    else:
        dirs = dirs.split(",")

    # Figure out where the rubin_sim_data is or is going
    data_dir = get_data_dir()
    if not os.path.isdir(data_dir):
        os.mkdir(data_dir)

    # Get dictionary of versions which are available on-disk
    versions = data_versions()
    if versions is None:
        versions = {}

    # ONLY check versions and return exit code
    if print_versions_only:
        print("Versions on disk currently // versions expected for this release:")
        mismatch_dict = {}
        match = True
        for k in file_dict:
            print(f"{k} : {versions.get(k, '')} // {file_dict[k]}")
            if versions.get(k, "") != file_dict[k]:
                match = False
                mismatch_dict[k] = False
        if match:
            print("Versions are in sync")
            return 0
        else:
            print("Versions do not match. ")
            print(f"{','.join([k for k in mismatch_dict])} are not matching.")
            return 1

    version_file = os.path.join(data_dir, "versions.txt")

    # See if base URL is alive
    url_base = url_base
    fail_message = f"Could not connect to {url_base}. Check site is up?"
    try:
        r = requests.get(url_base)
    except ConnectionError:
        print(fail_message)
        exit()
    if r.status_code != requests.codes.ok:
        print(fail_message)
        exit()

    # Now do downloading for "dirs"
    for key in dirs:
        filename = file_dict[key]
        path = os.path.join(data_dir, key)
        # Do some thinking to see if we should download new data for key
        download_this_dir = True
        if os.path.isdir(path):
            if force:
                # Remove and update, regardless
                rmtree(path)
                warnings.warn("Removed existing directory %s, downloading new copy" % path)
            elif not update:
                # Just see if it exists on-disk and keep it if it does
                warnings.warn("Directory %s already exists, skipping download" % path)
                download_this_dir = False
            else:
                # Update only if necessary
                if versions.get(key, "") == file_dict[key]:
                    download_this_dir = False
                else:
                    rmtree(path)
                    warnings.warn("Removed existing directory %s, downloading updated version" % path)
        if download_this_dir:
            # Download file
            url = url_base + filename
            print("Downloading file: %s" % url)
            # Stream and write in chunks (avoid large memory usage)
            r = requests.get(url, stream=True)
            file_size = int(r.headers.get("Content-Length", 0))
            if file_size < 245:
                warnings.warn(f"{url} file size unexpectedly small.")
            # Download this size chunk at a time; reasonable guess
            block_size = 512 * 512 * 10
            progress_bar = tqdm(total=file_size, unit="iB", unit_scale=True, disable=tdqm_disable)
            print(f"Writing to {os.path.join(data_dir, filename)}")
            with open(os.path.join(data_dir, filename), "wb") as f:
                for chunk in r.iter_content(chunk_size=block_size):
                    progress_bar.update(len(chunk))
                    f.write(chunk)
            progress_bar.close()
            # untar in place
            unpack_archive(os.path.join(data_dir, filename), data_dir)
            os.remove(os.path.join(data_dir, filename))
            versions[key] = file_dict[key]

    # Write out the new version info to the data directory
    with open(version_file, "w") as f:
        for key in versions:
            print(key + "," + versions[key], file=f)

    # Write a little table to stdout
    new_versions = data_versions()
    print("Current/updated data versions:")
    for k in new_versions:
        if len(k) <= 10:
            sep = "\t\t"
        else:
            sep = "\t"
        print(f"{k}{sep}{new_versions[k]}")
